import collections
import multiprocessing
import os
import re
import time
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import List

import pandas as pd
from COMPS.Data.Simulation import SimulationState
from tqdm import tqdm

from .base_aggregator import BaseAggregator
from .data_fetching import retrieve_data_for_simulation_aggregate
from simtools.DataAccess.Schema import Experiment
from simtools.Utilities.COMPSCache import COMPSCache
from simtools.Utilities.COMPSUtilities import COMPS_login
from simtools.Utilities.Experiments import retrieve_experiment


def monitor_queue(queue, total, aggregators, filename='aggregated_data_debu.h5'):
    print("Starting aggregator process")
    print([aggregator.uid for aggregator in aggregators])
    aggregate_output = pd.HDFStore(filename, mode='w')

    processed = 0
    with tqdm(total=total) as pbar:
        while True:
            data = queue.get()
            for aggregator in aggregators:
                if aggregator.uid in data:
                    pbar.set_description(f"Appending {processed} data to {aggregator.uid}")
                    aggregate_output.append(aggregator.uid, data[aggregator.uid])
            pbar.update()
            processed += 1
            if processed >= total:
                break


class AggregateDataManager:
    def __init__(self, aggregators: List[BaseAggregator], exp_list, max_sims=None, working_dir=None,
                 queue_function=monitor_queue):
        self.aggregators: List[BaseAggregator] = [aggregators] if type(aggregators) is BaseAggregator else aggregators
        self.experiments = set()
        self.max_sims = max_sims
        self.ignored_simulations = {}
        self.simulations = {}
        self.working_dir = working_dir or os.getcwd()
        self.verbose = True
        # Check if we are on SSMT
        ssmt_path_mapping = os.environ.get("COMPS_DATA_MAPPING", None)
        if ssmt_path_mapping:
            ssmt_path_mapping.lower().split(';')

        # Initial adding of experiments
        if exp_list:
            exp_list = exp_list if isinstance(exp_list, collections.Iterable) and not isinstance(exp_list,
                                                                                                 str) else [
                exp_list]
            for exp in exp_list: self.add_experiment(exp)

    def add_experiment(self, experiment: Experiment):
        if not isinstance(experiment, Experiment):
            experiment = retrieve_experiment(experiment)

        if experiment not in self.experiments:
            self.experiments.add(experiment)
            if experiment.location == "HPC":
                COMPS_login(experiment.endpoint)
                COMPSCache.load_experiment(experiment.exp_id)

            self.filter_simulations(experiment.simulations)

    def filter_simulations(self, simulations):
        if self.max_sims is not None:
            # If we already reached the maximum_simulation count -> exit
            if self.max_sims <= 0: return

            # Truncate the simulations to only add the needed ones
            simulations = simulations[:self.max_sims - len(self.simulations)]
            self.max_sims -= len(self.simulations)

        for s in simulations:
            if s.status != SimulationState.Succeeded:
                self.ignored_simulations[s.id] = s
            else:
                self.simulations[s.id] = s

    def add_aggregator(self, aggregator):
        # First check if we need to change the UID depending on other analyzers
        number = 1
        while any(a.uid == aggregator.uid for a in self.aggregators):
            # If we already suffixed with a number, remove it
            aggregator.uid = re.sub("\(\d+\)$", "", aggregator.uid)
            # Add the suffix
            aggregator.uid += "({})".format(number)
            number += 1

        aggregator.working_dir = aggregator.working_dir or self.working_dir

        # Then call the initialize method
        aggregator.initialize()

        self.aggregators.append(aggregator)

    def analyze(self):
        # Start the timer
        start_time = time.time()

        # If no analyzers -> quit
        if not all((self.aggregators, self.simulations)):
            print("No analyzers or experiments selected, exiting...")
            return False

        # Check if we are on SSMT
        ssmt_path_mapping = os.environ.get("COMPS_DATA_MAPPING", None)
        if ssmt_path_mapping:
            ssmt_path_mapping.lower().split(';')

        # If any of the analyzer needs the dir map, create it
        # Or if we are on SSMT
        if ssmt_path_mapping or any(a.need_dir_map for a in self.aggregators):
            # preload the global dir map
            from simtools.Utilities.SimulationDirectoryMap import SimulationDirectoryMap
            for experiment in self.experiments:
                SimulationDirectoryMap.preload_experiment(experiment)

        scount = len(self.simulations)

        if scount == 0 and self.verbose:
            print("No experiments/simulations for analysis.")
            return False

        with multiprocessing.Manager() as manager:
            q = manager.Queue(maxsize=64)
            write_process = multiprocessing.Process(target=self.queue_function, args=(q, scount, self.aggregators))
            write_process.start()
            with ProcessPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
                futures = []
                x = 0
                for simulation in tqdm(self.simulations.values()):
                    futures.append(
                        executor.submit(retrieve_data_for_simulation_aggregate, q, simulation, self.aggregators,
                                        ssmt_path_mapping))
                    if x > 1:
                        break
                    x += 1

                for f in as_completed(futures):
                    f.result()

            # nearly done. we need to check queue and monitor until it is empty
            while not q.empty():
                time.sleep(5)
            write_process.join()
        return True
