import collections
import itertools
import os
import re
import time
from typing import List
from dask.delayed import delayed
import pandas as pd

import dask.dataframe as dd
from COMPS.Data.Simulation import SimulationState
from dask.diagnostics import ProgressBar
from tqdm import tqdm

from .step1_aggregate import SpatialAggregator, GMEstablishmentAggregator
from .base_aggregator import BaseAggregator
from simtools.DataAccess.Schema import Experiment
from simtools.Utilities.COMPSCache import COMPSCache
from simtools.Utilities.COMPSUtilities import COMPS_login
from simtools.Utilities.Experiments import retrieve_experiment
from simtools.Analysis.DataRetrievalProcess import retrieve_SSMT_files, retrieve_COMPS_AM_files
from simtools.Analysis.OutputParser import SimulationOutputParser


def retrieve_data_for_simulation_aggregate_dask(simulation, aggregator, path_mapping):
    # Filter first and get the filenames from filtered analysis
    filtered_aggregator = [a for a in [aggregator] if a.filter(simulation)]
    filenames = set(itertools.chain(*(a.filenames for a in filtered_aggregator)))

    # We dont have anything to do :)
    if not filtered_aggregator:
        return {}

    # The byte_arrays will associate filename with content
    byte_arrays = {}

    # Retrieval for SSMT
    if path_mapping:
        byte_arrays = retrieve_SSMT_files(simulation, filenames, path_mapping)

    # Retrieval for normal HPC Asset Management
    elif simulation.experiment.location == "HPC":
        # global comps cache for local runs
        byte_arrays = retrieve_COMPS_AM_files(simulation, filenames)

    # Retrieval for local file
    else:
        for filename in filenames:
            path = os.path.join(simulation.get_path(), filename)
            with open(path, 'rb') as output_file:
                byte_arrays[filename] = output_file.read()

    # Selected data will be a dict with analyzer.uid => data
    selected_data = {}
    for aggregator in filtered_aggregator:
        # If the analyzer needs the parsed data, parse
        if aggregator.parse:
            data = {filename: SimulationOutputParser.parse(filename, content)
                    for filename, content in byte_arrays.items()}
        else:
            data = byte_arrays

        # Retrieve the selected data for the given analyzer
        selected_data[aggregator.uid] = aggregator.select_simulation_data(data, simulation)
    # wait until the queue has room
    return selected_data


@delayed
def load_aggregator_to_dask(simulation, aggregator: BaseAggregator, ssmt_path_mapping):
    return retrieve_data_for_simulation_aggregate_dask(simulation, aggregator, ssmt_path_mapping)[aggregator.uid]


class DaskAggregateDataManager:
    def __init__(self, aggregators: List[BaseAggregator], exp_list, max_sims=None, working_dir=None):
        self.aggregators: List[BaseAggregator] = [aggregators] if type(aggregators) is BaseAggregator else aggregators
        self.experiments = set()
        self.max_sims = max_sims
        self.ignored_simulations = {}
        self.simulations = {}
        self.working_dir = working_dir or os.getcwd()
        self.verbose = True
        # Check if we are on SSMT
        ssmt_path_mapping = os.environ.get("COMPS_DATA_MAPPING", None)
        if ssmt_path_mapping:
            ssmt_path_mapping.lower().split(';')

        # Initial adding of experiments
        if exp_list:
            exp_list = exp_list if isinstance(exp_list, collections.Iterable) and not isinstance(exp_list,
                                                                                                 str) else [
                exp_list]
            for exp in exp_list: self.add_experiment(exp)

    def add_experiment(self, experiment: Experiment):
        if not isinstance(experiment, Experiment):
            experiment = retrieve_experiment(experiment)

        if experiment not in self.experiments:
            self.experiments.add(experiment)
            if experiment.location == "HPC":
                COMPS_login(experiment.endpoint)
                COMPSCache.load_experiment(experiment.exp_id)

            self.filter_simulations(experiment.simulations)

    def filter_simulations(self, simulations):
        if self.max_sims is not None:
            # If we already reached the maximum_simulation count -> exit
            if self.max_sims <= 0: return

            # Truncate the simulations to only add the needed ones
            simulations = simulations[:self.max_sims - len(self.simulations)]
            self.max_sims -= len(self.simulations)

        for s in simulations:
            if s.status != SimulationState.Succeeded:
                self.ignored_simulations[s.id] = s
            else:
                self.simulations[s.id] = s

    def add_aggregator(self, aggregator):
        # First check if we need to change the UID depending on other analyzers
        number = 1
        while any(a.uid == aggregator.uid for a in self.aggregators):
            # If we already suffixed with a number, remove it
            aggregator.uid = re.sub("\(\d+\)$", "", aggregator.uid)
            # Add the suffix
            aggregator.uid += "({})".format(number)
            number += 1

        aggregator.working_dir = aggregator.working_dir or self.working_dir

        # Then call the initialize method
        aggregator.initialize()

        self.aggregators.append(aggregator)

    def analyze(self):
        """
        Returns the load into dask dataframes
        :return:
        """
        # Start the timer
        start_time = time.time()

        # If no analyzers -> quit
        if not all((self.aggregators, self.simulations)):
            print("No analyzers or experiments selected, exiting...")
            return False

        # Check if we are on SSMT
        ssmt_path_mapping = os.environ.get("COMPS_DATA_MAPPING", None)
        if ssmt_path_mapping:
            ssmt_path_mapping.lower().split(';')

        # If any of the analyzer needs the dir map, create it
        # Or if we are on SSMT
        if ssmt_path_mapping or any(a.need_dir_map for a in self.aggregators):
            # preload the global dir map
            from simtools.Utilities.SimulationDirectoryMap import SimulationDirectoryMap
            for experiment in self.experiments:
                SimulationDirectoryMap.preload_experiment(experiment)

        scount = len(self.simulations)

        if scount == 0 and self.verbose:
            print("No experiments/simulations for analysis.")
            return False

        # compile array of dask delays into each list. This should reduce it to a way where we can find all the data we
        # want but delay its actual load until warranted
        results = dict()
        for simulation in tqdm(self.simulations.values()):
            for agg in self.aggregators:
                results[agg.uid] = load_aggregator_to_dask(simulation, agg, ssmt_path_mapping)
        # convert load functions into dask data frames
        for agg, result in results.items():
            results[agg] = dd.from_delayed(result)
        return results


if __name__ == "__main__":

    experiments = {
        # "VC_only": "36f37682-158a-e911-a2c1-c4346bcb1555",
        # "VC_and_GM": "35f37682-158a-e911-a2c1-c4346bcb1555",
        # "No_interventions": "e264b5ee-be8b-e911-a2c1-c4346bcb1555",
        # "GM_only": "4880820c-8991-e911-a2c1-c4346bcb1555",
        "VC_and_GM_vary_release": "b8bb4828-a599-e911-a2c1-c4346bcb1555"
    }
    # The channels we want to regulate
    channels = ['Prevalence', 'Daily_EIR', 'Adult_Vectors', 'New_Clinical_Cases']
    # our sweep vars
    sweep_vars = ['Run_Number', 'Release_number', 'Copy_To_Likelihood', 'ITN_Coverage', 'Transmission_To_Human']

    # Loop through our experiments
    for expt_name, exp_id in experiments.items():
        # Define our aggregators
        aggregators = [SpatialAggregator(exp_name=expt_name, spatial_channels=channels, sweep_variables=sweep_vars),
                       GMEstablishmentAggregator(exp_name=expt_name, sweep_variables=sweep_vars)]
        # And define our aggregate data manager
        am = DaskAggregateDataManager(exp_list=exp_id, aggregators=aggregators)
        # get a dict that is our dask data frames
        result = am.analyze()

        group_vars = ['Release_number', 'Copy_To_Likelihood', 'ITN_Coverage', 'Transmission_To_Human', 'NodeID', 'Time']

        # now do our analysis on the data by computing the group by of our data load. Of course, this is actually
        # is when our real data load happens as well
        for agg, values in result.items():
            with ProgressBar():
                result[agg] = values.groupby(group_vars)[channels].mean().compute()

        # merge the group bys for final data form
        final_ds = pd.merge(result['SpatialAggregator'], result['GMEstablishmentAggregator'], how='left',
                            on=group_vars)
        final_ds.to_csv('final.csv', index=False)
