import os, sys
import pandas as pd

sys.path.append(os.path.dirname(__file__))

from aggregate_data_manager import AggregateDataManager
from base_aggregator import BaseAggregator

from simtools.Utilities.General import init_logging

logger = init_logging('AggregateData')

column_types = dict(Time='uint16',
                    time='uint16',
                    New_Clinical_Cases='uint8',
                    Adult_Vectors='uint32',
                    Run_Number='uint16',
                    Release_number='uint16',
                    Node='uint8',
                    node='uint8')


def optimize_dataframe(simdata):
    """
    Set the data to the optimal size for their types. For examples, we know this experiment does not have
    more than 255 nodes so we should assign the NodeId columns to an uint8 since it takes less memory than the
    default uint64

    :param simdata:
    :return:
    """
    for column, item_type in column_types.items():
        if column in simdata.columns:
            simdata[column] = simdata[column].astype(item_type)
    return simdata


def construct_spatial_output_df(rawdata, channel, timesteps=None):
    if timesteps is None:
        timesteps = []
    n_nodes = rawdata['n_nodes']
    n_tstep = rawdata['n_tstep']
    if 'start' in rawdata:
        start = rawdata['start']
        interval = rawdata['interval']
    else:
        start, interval = 0, 1
    nodeids = rawdata['nodeids']
    data = rawdata['data']

    all_timesteps = range(start, (start + n_tstep) * interval, interval)

    df = pd.DataFrame({channel: [item for sublist in data for item in sublist],
                       'time': [item for sublist in [[x] * n_nodes for x in all_timesteps] for item in sublist],
                       'node': [item for sublist in [list(nodeids) * len(all_timesteps)] for item in sublist]})
    if not timesteps:
        return df

    timesteps = sorted(list(set(all_timesteps).intersection(timesteps)))
    return df[df['time'].isin(timesteps)]


class SpatialAggregator(BaseAggregator):

    def __init__(self, exp_name, spatial_channels, sweep_variables, working_dir='.'):
        input_files = ['output/SpatialReportMalariaFiltered_%s.bin' % x for x in spatial_channels]
        super(SpatialAggregator, self).__init__(working_dir=working_dir, filenames=input_files)

        self.exp_name = exp_name
        self.sweep_variables = sweep_variables
        self.spatial_channels = spatial_channels
        self.output_fname = os.path.join(self.working_dir, "%s_spatial_data.csv" % self.exp_name)
        self.cache = True

    def select_simulation_data(self, data, simulation):

        simdata = construct_spatial_output_df(
            data['output/SpatialReportMalariaFiltered_%s.bin' % self.spatial_channels[0]], self.spatial_channels[0])
        if len(self.spatial_channels) > 1:
            for ch in self.spatial_channels[1:]:
                simdata = pd.merge(left=simdata,
                                   right=construct_spatial_output_df(
                                       data['output/SpatialReportMalariaFiltered_%s.bin' % ch], ch),
                                   on=['time', 'node'])

        for sweep_var in self.sweep_variables:
            if sweep_var in simulation.tags.keys():
                simdata[sweep_var] = simulation.tags[sweep_var]
            else:
                simdata[sweep_var] = 0

        # optimize df types
        simdata = optimize_dataframe(simdata)
        # Update Columns names to match data from GM Analyzer
        simdata.rename(columns={'node': 'NodeID', 'time': 'Time'}, inplace=True)

        # return out_path
        return simdata


class GMEstablishmentAggregator(BaseAggregator):

    def __init__(self, exp_name, sweep_variables, working_dir='.'):
        input_files = ['output/ReportVectorGenetics_gambiae_Female_SPECIFIC_GENOME.csv']
        super(GMEstablishmentAggregator, self).__init__(working_dir=working_dir, filenames=input_files)

        self.exp_name = exp_name
        self.sweep_variables = sweep_variables
        self.output_fname = os.path.join(self.working_dir, "%s_establishment_rates.csv" % self.exp_name)
        self.cache = True

    def select_simulation_data(self, data, simulation):

        simdata = data['output/ReportVectorGenetics_gambiae_Female_SPECIFIC_GENOME.csv']
        genomes = simdata['Genome'].unique()
        simdata = simdata.pivot_table('VectorPopulation', ['Time', 'NodeID'], 'Genome').reset_index()
        simdata['sum'] = [0] * len(simdata)
        for genome in genomes:
            simdata['sum'] += simdata[genome]
        for genome in genomes:
            simdata[genome] = simdata[genome] / simdata['sum']
            simdata = simdata.rename(columns={'X-a0:X-a0': 'WT'})
        genomes = genomes.tolist()
        genomes.remove('X-a0:X-a0')
        simdata['GM'] = [0] * len(simdata)
        for genome in genomes:
            simdata['GM'] += simdata[genome]
        simdata = simdata[['Time', 'NodeID', 'GM', 'WT']]

        for sweep_var in self.sweep_variables:
            if sweep_var in simulation.tags.keys():
                simdata[sweep_var] = simulation.tags[sweep_var]
            else:
                simdata[sweep_var] = 0

        simdata = optimize_dataframe(simdata)
        return simdata


if __name__ == "__main__":

    experiments = {
        # "VC_only": "36f37682-158a-e911-a2c1-c4346bcb1555",
        # "VC_and_GM": "35f37682-158a-e911-a2c1-c4346bcb1555",
        # "No_interventions": "e264b5ee-be8b-e911-a2c1-c4346bcb1555",
        # "GM_only": "4880820c-8991-e911-a2c1-c4346bcb1555",
        "VC_and_GM_vary_release": "b8bb4828-a599-e911-a2c1-c4346bcb1555"
    }
    # The channels we want to regulate
    channels = ['Prevalence', 'Daily_EIR', 'Adult_Vectors', 'New_Clinical_Cases']
    # our sweep vars
    sweep_vars = ['Run_Number', 'Release_number', 'Copy_To_Likelihood', 'ITN_Coverage', 'Transmission_To_Human']

    # Loop through our experiments
    for expt_name, exp_id in experiments.items():
        # Define our aggregators
        aggregators = [SpatialAggregator(exp_name=expt_name, spatial_channels=channels, sweep_variables=sweep_vars),
                       GMEstablishmentAggregator(exp_name=expt_name, sweep_variables=sweep_vars)]
        # And define our aggregate data manager
        am = AggregateDataManager(exp_list=exp_id, aggregators=aggregators)
        am.analyze()
