import copy
import os
import pandas as pd
from helper_functions import optimize_dataframe
from simtools.Analysis.AnalyzeManager import AnalyzeManager
from simtools.Analysis.BaseAnalyzers import BaseAnalyzer
from simtools.SetupParser import SetupParser


class GMEstablishmentAnalyzer(BaseAnalyzer):

    def __init__(self, exp_name, sweep_variables, working_dir='.'):
        super(GMEstablishmentAnalyzer, self).__init__(working_dir=working_dir,
                                                      filenames=['output/'
                                                                 'ReportVectorGenetics_gambiae_Female_SPECIFIC_GENOME.csv']
                                                      )

        self.exp_name = exp_name
        self.sweep_variables = sweep_variables
        self.group_vars = list(sorted(copy.deepcopy(self.sweep_variables) + ['time', 'node']))
        self.group_vars.remove('Run_Number')
        self.output_fname = os.path.join(self.working_dir, "%s_establishment_rates.h5" % self.exp_name)

    def select_simulation_data(self, data, simulation):

        simdata = data['output/ReportVectorGenetics_gambiae_Female_SPECIFIC_GENOME.csv']
        genomes = simdata['Genome'].unique()
        simdata = simdata.pivot_table('VectorPopulation', ['Time', 'NodeID'], 'Genome').reset_index()
        simdata['sum'] = [0] * len(simdata)
        for genome in genomes:
            simdata['sum'] += simdata[genome]
        for genome in genomes:
            simdata[genome] = simdata[genome] / simdata['sum']
            simdata = simdata.rename(columns={'X-a0:X-a0': 'WT'})
        genomes = genomes.tolist()
        genomes.remove('X-a0:X-a0')
        simdata['GM'] = [0] * len(simdata)
        for genome in genomes:
            simdata['GM'] += simdata[genome]
        simdata = simdata[['Time', 'NodeID', 'GM', 'WT']]
        simdata.rename(columns={"Time": "time", "NodeID": "node"}, inplace=True)

        for sweep_var in self.sweep_variables:
            if sweep_var in simulation.tags.keys():
                simdata[sweep_var] = simulation.tags[sweep_var]
            else:
                simdata[sweep_var] = 0

        # we don't want index with aggregation of dask
        simdata.reset_index(drop=True)
        return optimize_dataframe(simdata)

    def finalize(self, data: pd.DataFrame):
        data = data.groupby(self.group_vars, observed=True)['GM'].mean()
        data = data.reset_index()
        data = optimize_dataframe(data)
        if os.path.exists(self.output_fname):
            print(pd.read_hdf(self.output_fname, self.uid, stop=100).dtypes)
        print(data.dtypes)
        data.to_hdf(self.output_fname, self.uid, mode='a' if os.path.exists(self.output_fname) else 'w',
                    append=os.path.exists(self.output_fname), format='table', complib='blosc', complevel=3)


if __name__ == "__main__":

    SetupParser.default_block = 'HPC'
    SetupParser.init()

    experiments = {
        "GM_test": "575bacfe-39c9-e911-a2c1-c4346bcb1555"
    }

    sweep_vars = ['Run_Number', 'Release_Number', 'Copy_To_Likelihood',
                  'ITN_Coverage', 'Transmission_To_Human',
                  'Start_Day', 'Num_Nodes']

    for expt_name, exp_id in experiments.items():
        am = AnalyzeManager(exp_list=exp_id,
                            analyzers=[
                                GMEstablishmentAnalyzer(exp_name='GM_release',
                                                        working_dir=".",
                                                        sweep_variables=sweep_vars
                                                        )
                            ],
                            force_analyze=False)

        print(am.experiments)
        am.analyze()
