import os
import sys

from simtools.Utilities.COMPSCache import COMPSCache
from simtools.Utilities.COMPSUtilities import COMPS_login, get_asset_files_for_simulation_id
from simtools.Utilities.RetryDecorator import retry

sys.path.append(os.path.dirname(__file__))
from dask.diagnostics import ProgressBar
from tqdm import tqdm
import functools
import itertools
import shutil
import backoff as backoff
import dask.dataframe as dd
from dask.distributed import Client, LocalCluster
from collections import defaultdict
from dask.delayed import delayed
from simtools.Analysis.DataRetrievalProcess import retrieve_SSMT_files, retrieve_COMPS_AM_files
from simtools.Utilities import pluralize, on_off
from simtools.Analysis.OutputParser import SimulationOutputParser
from simtools.Analysis.AnalyzeManager import AnalyzeManager
from spatial_data_analyzer import SpatialAnalyzer
from GM_data_analyzer import GMEstablishmentAnalyzer


@retry(ConnectionError, tries=100, delay=3, backoff=2)
def retrieve_COMPS_AM_files(endpoint, simulation_id, filenames):
    byte_arrays = {}

    # Login and retrieve the COMPS simulation
    COMPS_login(endpoint)
    COMPS_simulation = COMPSCache.simulation(simulation_id)

    # Separate the files in assets (for those with Assets in the path) and transient (for the others)
    assets = [path for path in filenames if path.lower().startswith("assets")]
    transient = [path for path in filenames if not path.lower().startswith("assets")]

    # Retrieve and store
    if transient:
        byte_arrays.update(dict(zip(transient, COMPS_simulation.retrieve_output_files(paths=transient))))

    if assets:
        byte_arrays.update(get_asset_files_for_simulation_id(simulation_id, paths=assets, remove_prefix='Assets'))

    data = {filename: SimulationOutputParser.parse(filename, content)
            for filename, content in byte_arrays.items()}

    return data


def retrieve_data(analyzers, simulation):
    selected_data = dict()
    for analyzer in analyzers:
        selected_data[analyzer.uid] = do_select(analyzer, simulation.experiment.endpoint, simulation.id,
                                                simulation.tags)
    return selected_data


@delayed(pure=True)
def do_select(analyzer, endpoint, simulation_id, tags):
    selected_data = retrieve_COMPS_AM_files(endpoint, simulation_id, analyzer.filenames)
    return analyzer.select_simulation_data_dask(selected_data, simulation_id, tags)


class DaskAnalyzeManager(AnalyzeManager):
    def analyze(self, client):
        # Clear the cache
        self.cache = self.initialize_cache(shards=self.max_threads)

        # Check if we are on SSMT
        ssmt_path_mapping = os.environ.get("COMPS_DATA_MAPPING", None)
        if ssmt_path_mapping:
            ssmt_path_mapping.lower().split(';')

        # If any of the analyzer needs the dir map, create it
        # Or if we are on SSMT
        if ssmt_path_mapping or any(a.need_dir_map for a in self.analyzers):
            # preload the global dir map
            from simtools.Utilities.SimulationDirectoryMap import SimulationDirectoryMap
            for experiment in self.experiments:
                SimulationDirectoryMap.preload_experiment(experiment)

        # Run the per experiment on the analyzers
        # for exp in self.experiments:
        #     for a in self.analyzers:
        #         a.per_experiment(exp)

        # Display some info

        scount = len(self.simulations)
        max_threads = min(self.max_threads, scount if scount != 0 else 1)

        if self.verbose:
            print("Analyze Manager")
            print(" | {} simulation{} - {} experiment{}"
                  .format(scount, pluralize(scount), len(self.experiments), pluralize(self.experiments)))
            print(" | force_analyze is {} and {} simulation{} ignored"
                  .format(on_off(self.force_analyze), len(self.ignored_simulations),
                          pluralize(self.ignored_simulations)))
            print(" | Analyzer{}: ".format(pluralize(self.analyzers)))
            for a in self.analyzers:
                print(" |  - {} (Directory map: {} / File parsing: {} / Use cache: {})"
                      .format(a.uid, on_off(a.need_dir_map), on_off(a.parse), on_off(hasattr(a, "cache"))))
            print(" | Pool of {} analyzing processes".format(max_threads))

        if scount == 0 and self.verbose:
            print("No experiments/simulations for analysis.")
            return False

        retrieve_call = functools.partial(retrieve_data, self.analyzers)
        print('Scattering data across workers')
        delayed_result_dicts = map(retrieve_call, self.simulations.values())

        # merge dicts into one dict with two arrays
        delayed_dict = defaultdict(list)
        for d in tqdm(delayed_result_dicts, desc="Building Dask Delayed Select Statement"):
            [delayed_dict[k].append(v) for k, v in d.items()]

        results = dict()
        # now produce dask dataframes from them
        # convert load functions into dask data frames
        for analyzer in self.analyzers:
            with ProgressBar():
                print('Loading from_delayed')
                results[analyzer.uid] = dd.from_delayed(delayed_dict[analyzer.uid])
                # now finalize
                results[analyzer.uid] = analyzer.finalize(results[analyzer.uid])


if __name__ == "__main__":
    cluster = LocalCluster(dashboard_address='0.0.0.0:8787')
    client = Client(cluster)
    experiments = {
        # "VC_only": "36f37682-158a-e911-a2c1-c4346bcb1555",
        # "VC_and_GM": "35f37682-158a-e911-a2c1-c4346bcb1555",
        # "No_interventions": "e264b5ee-be8b-e911-a2c1-c4346bcb1555",
        # "GM_only": "4880820c-8991-e911-a2c1-c4346bcb1555",
        # "VC_and_GM_vary_release": "b8bb4828-a599-e911-a2c1-c4346bcb1555",
        "VC_GM_test": "575bacfe-39c9-e911-a2c1-c4346bcb1555"
    }
    channels = ['Prevalence', 'Daily_EIR', 'Adult_Vectors', 'New_Clinical_Cases']
    sweep_vars = ['Run_Number', 'Release_Number', 'Copy_To_Likelihood',
                  'ITN_Coverage', 'Transmission_To_Human',
                  'Start_Day', 'Num_Nodes']

    for expt_name, exp_id in experiments.items():
        am = DaskAnalyzeManager(exp_list=exp_id,
                                analyzers=
                                [SpatialAnalyzer(exp_name=expt_name,
                                                 spatial_channels=channels,
                                                 sweep_variables=sweep_vars
                                                 ),
                                 GMEstablishmentAnalyzer(exp_name=expt_name,
                                                         sweep_variables=sweep_vars
                                                         )
                                 ]
                                )
        print(am.experiments)
        am.analyze(client)
