import copy
import os
import pandas as pd
from helper_functions import optimize_dataframe
from simtools.Analysis.AnalyzeManager import AnalyzeManager
from simtools.Analysis.BaseAnalyzers import BaseAnalyzer
from simtools.SetupParser import SetupParser
from spatial_output_dataframe import construct_spatial_output_df


class SpatialAnalyzer(BaseAnalyzer):

    def __init__(self, exp_name, spatial_channels, sweep_variables, working_dir='.'):
        super(SpatialAnalyzer, self).__init__(working_dir=working_dir,
                                              filenames=['output/SpatialReportMalariaFiltered_%s.bin' % x for x in
                                                         spatial_channels]
                                              )

        self.exp_name = exp_name
        self.sweep_variables = sweep_variables
        self.group_vars = list(sorted(copy.deepcopy(self.sweep_variables) + ['time', 'node']))
        self.group_vars.remove('Run_Number')
        self.spatial_channels = spatial_channels
        self.output_fname = os.path.join(self.working_dir, "%s_spatial_data.h5" % self.exp_name)

    def select_simulation_data(self, data, simulation):

        simdata = construct_spatial_output_df(
            data['output/SpatialReportMalariaFiltered_%s.bin' % self.spatial_channels[0]], self.spatial_channels[0])
        if len(self.spatial_channels) > 1:
            for ch in self.spatial_channels[1:]:
                simdata = pd.merge(left=simdata,
                                   right=construct_spatial_output_df(
                                       data['output/SpatialReportMalariaFiltered_%s.bin' % ch], ch),
                                   on=['time', 'node'])

        for sweep_var in self.sweep_variables:
            if sweep_var in simulation.tags.keys():
                simdata[sweep_var] = simulation.tags[sweep_var]
            else:
                simdata[sweep_var] = 0

        return optimize_dataframe(simdata)

    def finalize(self, data: pd.DataFrame):
        print("before grouping " + str(data.dtypes))
        data = data.groupby(self.group_vars, observed=True)[self.spatial_channels].mean()
        data = data.reset_index()
        data = optimize_dataframe(data)
        if os.path.exists(self.output_fname):
            print("Existing dtypes: " + str(pd.read_hdf(self.output_fname, self.uid, stop=100).dtypes))
        print("Final dtypes: " + str(data.dtypes))
        data.to_hdf(self.output_fname, self.uid, mode='a' if os.path.exists(self.output_fname) else 'w',
                    append=os.path.exists(self.output_fname), format='table', complib='blosc', complevel=3)


if __name__ == "__main__":

    SetupParser.default_block = 'HPC'
    SetupParser.init()

    out_dir = '.'

    experiments = {
        "GM_test": "575bacfe-39c9-e911-a2c1-c4346bcb1555"
    }
    sweep_vars = ['Run_Number', 'Release_Number', 'Copy_To_Likelihood',
                  'ITN_Coverage', 'Transmission_To_Human',
                  'Start_Day', 'Num_Nodes']

    for expt_name, exp_id in experiments.items():
        am = AnalyzeManager(exp_list=exp_id,
                            analyzers=[
                                SpatialAnalyzer(exp_name='GM_release',
                                                working_dir=out_dir,
                                                sweep_variables=sweep_vars,
                                                spatial_channels=['Prevalence', 'Daily_EIR', 'Adult_Vectors'],
                                                )
                            ],
                            force_analyze=False)

        print(am.experiments)
        am.analyze()
